/*
  parse (...) :
       1. Reads maximal flow problem in extended DIMACS format.
       2. Prepares internal data representation.
       
   types: 'arc' and 'node' must be predefined

   type arc  must contain fields 'head', 'rev', 'resCap'

   typedef 
     struct arcSt
       {
         long             resCap;     .. residual capasity
         struct nodeSt   *head;     .. head node 
         struct arcSt    *rev;   .. opposite arc 
         ..................
       }
   arc;

   type   node   must contain the field 'first': 

   typedef
     struct nodeSt
       {
          arcSt        *first;    ..  first outgoing arc 
          ....................
       }
    node;
*/

/* ----------------------------------------------------------------- */

#include <iostream>
#include <iomanip>

void exit_on_error(const string err_msg)
{
  cout << err_msg << endl;
  exit(1);
}

static int _N, _M;
static node *p_nodes;
static arc *p_arcs;
static long *p_acap;
static long *p_arc_tail;
static long *p_arc_first;

void hi_pr_init(int n, int m)
{
  _N = n;  _M = 2*m;
 /* allocating memory for  'nodes', 'arcs'  and internal arrays */
  p_nodes    = (node*) calloc ( _N+2, sizeof(node) );
  p_arcs     = (arc*)  calloc ( 2*_M+1, sizeof(arc) );
  p_arc_tail = (long*) calloc ( 2*_M,   sizeof(long) ); 
  p_arc_first= (long*) calloc ( _N+2, sizeof(long) );
  p_acap     = (long*) calloc ( 2*_M, sizeof(long) );
  buckets = (bucket*) calloc ( _N+2, sizeof (bucket) );
  /* arc_first [ 0 .. n+1 ] = 0 - initialized by calloc */

  // allocating memory for the scc graph
  scc_nodes    = (scc_node*) calloc ( _N+1, sizeof(scc_node) );
  scc_arcs     = (scc_arc*)  calloc ( 2*_M+1, sizeof(scc_arc) );

  if ( p_nodes == NULL || p_arcs == NULL || 
       p_arc_first == NULL || p_arc_tail == NULL || buckets == NULL ||
       scc_nodes == NULL || scc_arcs == NULL)
    /* memory is not allocated */
    { exit_on_error("hi_pr_init: allocation memory error"); }
}

void hi_pr_free() 
{
  free(p_nodes);
  free(p_arcs);
  free(p_arc_tail);
  free(p_arc_first);
  free(p_acap);
  free(buckets);
  free(scc_nodes);
  free(scc_arcs);
}


int parse_edgelist(Edgelist& G, long *n_ad, long *m_ad, node **nodes_ad, 
		   arc **arcs_ad, cType **cap_ad, node **source_ad, 
		   node **sink_ad, long *node_min_ad)
{

#define MAXLINE       100	/* max line length in the input file */
#define ARC_FIELDS      3	/* no of fields in arc line  */
#define NODE_FIELDS     2	/* no of fields in node line  */
#define P_FIELDS        3       /* no of fields in problem line */
#define PROBLEM_TYPE "max"      /* name of problem type*/


long    n,                      /* internal number of nodes */
        node_min=0,             /* minimal no of node  */
        node_max=0,             /* maximal no of nodes */
       *arc_first=NULL,         /* internal array for holding
                                     - node degree
                                     - position of the first outgoing arc */
       *arc_tail=NULL,          /* internal array: tails of the arcs */
        source=0,               /* no of the source */
        sink=0,                 /* no of the sink */
        /* temporary variables carrying no of nodes */
        head, tail, i;

long    m,                      /* internal number of arcs */
        /* temporary variables carrying no of arcs */
        last, arc_num, arc_new_num;

node    *nodes=NULL,            /* pointer to the node structure */
        *head_p,
        *ndp;

arc     *arcs=NULL,             /* pointer to the arc structure */
        *arc_current=NULL,
        *arc_new,
        *arc_tmp;

long    *acap=NULL,             /* array of capasities */
        cap;                    /* capasity of the current arc */

long    no_lines=0,             /* no of current input line */
        no_plines=0,            /* no of problem-lines */
        no_nslines=0,           /* no of node-source-lines */
        no_nklines=0,           /* no of node-source-lines */
        no_alines=0,            /* no of arc-lines */
        pos_current=0;          /* 2*no_alines */

char    in_line[MAXLINE],       /* for reading input line */
        pr_type[4],             /* for reading type of the problem */
        nd;                     /* source (s) or sink (t) */

int     k,                      /* temporary */
        err_no;                 /* no of detected error */

int i_edges;
long tmp;

/* -------------- error numbers & error messages ---------------- */
#define EN1   0
#define EN2   1
#define EN3   2
#define EN4   3
#define EN6   4
#define EN10  5
#define EN7   6
#define EN8   7
#define EN9   8
#define EN11  9
#define EN12 10
#define EN13 11
#define EN14 12
#define EN16 13
#define EN15 14
#define EN17 15
#define EN18 16
#define EN21 17
#define EN19 18
#define EN20 19
#define EN22 20

const string err_message[] = 
  { 
/* 0*/    "more than one problem line.",
/* 1*/    "wrong number of parameters in the problem line.",
/* 2*/    "it is not a Max Flow problem line.",
/* 3*/    "bad value of a parameter in the problem line.",
/* 4*/    "can't obtain enough memory to solve this problem.",
/* 5*/    "more than one line with the problem name.",
/* 6*/    "can't read problem name.",
/* 7*/    "problem description must be before node description.",
/* 8*/    "this parser doesn't support multiply sources and sinks.",
/* 9*/    "wrong number of parameters in the node line.",
/*10*/    "wrong value of parameters in the node line.",
/*11*/    " ",
/*12*/    "source and sink descriptions must be before arc descriptions.",
/*13*/    "too many arcs in the input.",
/*14*/    "wrong number of parameters in the arc line.",
/*15*/    "wrong value of parameters in the arc line.",
/*16*/    "unknown line type in the input.",
/*17*/    "reading error.",
/*18*/    "not enough arcs in the input.",
/*19*/    "source or sink doesn't have incident arcs.",
/*20*/    "can't read anything from the input file."
  };
/* --------------------------------------------------------------- */

/* The main loop:
        -  reads the line of the input,
        -  analises its type,
        -  checks correctness of parameters,
        -  puts data to the arrays,
        -  does service functions
*/

/* initialize structures */
 n = G.num_nodes();
 m = 2*G.num_edges(); /* double for anti-parallel edges */

 if (n > _N || m > _M)
   exit_on_error("parser: graph larger than previously allocated.");
 if ( n <= 0  || m <= 0 )
   /*wrong value of no of arcs or nodes*/
   { err_no = EN4; exit_on_error(err_message[err_no]); }
 source = G.get_source();
 sink = G.get_sink();
 if ( source < 0 || source > n )
   /* wrong value of node */
   { err_no = EN12; exit_on_error(err_message[err_no]); }
 if ( sink   < 0 || sink > n )
   /* wrong value of node */
   { err_no = EN12; exit_on_error(err_message[err_no]); }
 node_max = 0;
 node_min = n;

 /* allocating memory for  'nodes', 'arcs'  and internal arrays */
 nodes    = p_nodes;
 arcs     = p_arcs;
 arc_tail = p_arc_tail;
 arc_first= p_arc_first;
 acap     = p_acap;
 
 /* arc_first [ 0 .. n+1 ] = 0 - now requires explicit initialization */
 for (i = 0; i <= n+1; i++)
   arc_first[i] = 0;

 if ( nodes == NULL || arcs == NULL || 
      arc_first == NULL || arc_tail == NULL )
   /* memory is not allocated */
   { err_no = EN6; exit_on_error(err_message[err_no]); }
 
 /* setting pointer to the first arc */
 arc_current = arcs;

 Edgelist::edge_iterator edge_it;
 for (edge_it = G.begin(); edge_it != G.end(); ++edge_it)
   {
     Edge edge = *edge_it;
     no_lines ++;
     
     if ( no_alines >= m/2 )
       /*too many arcs on input*/
       { err_no = EN16; exit_on_error(err_message[err_no]); }
     tail = (long) edge.tail;
     head = (long) edge.head;
     cap = (long) edge.cap;
     if ( tail < 0  ||  tail > n  ||
	  head < 0  ||  head > n  
	  )
       /* wrong value of nodes */
       { err_no = EN17; exit_on_error(err_message[err_no]); }
     
     // undirected version - add anti-parallel edges
     for (i_edges=0; i_edges < 2; ++i_edges)
       {
	 // printf("Parser: adding edge (%ld, %ld, %ld)\n", tail, head, cap);

	 /* no of arcs incident to node i is stored in arc_first[i+1] */
	 arc_first[tail + 1] ++; 
	 arc_first[head + 1] ++;
	 
	 /* storing information about the arc */
	 arc_tail[pos_current]        = tail;
	 arc_tail[pos_current+1]      = head;
	 arc_current       -> head    = nodes + head;
	 arc_current       -> resCap    = cap;
	 arc_current       -> rev  = arc_current + 1;
	 ( arc_current + 1 ) -> head    = nodes + tail;
	 ( arc_current + 1 ) -> resCap    = 0;
	 ( arc_current + 1 ) -> rev  = arc_current;
	 
	 arc_current += 2;
	 pos_current += 2;
	 // added: switch roles of head and tail 
	 tmp = head;
	 head = tail;
	 tail = tmp; 	
       } /* end for */
     /* searching minimum and maximum node */
     if ( head < node_min ) node_min = head;
     if ( tail < node_min ) node_min = tail;
     if ( head > node_max ) node_max = head;
     if ( tail > node_max ) node_max = tail;
     no_alines   ++;
  }     /* end of while loop */

if ( no_lines == 0 ) /* empty input */
  { err_no = EN22; exit_on_error(err_message[err_no]); } 

if ( no_alines < m/2 ) /* not enough arcs */
  { err_no = EN19; exit_on_error(err_message[err_no]); } 

/********** ordering arcs - linear time algorithm ***********/

/* first arc from the first node */
( nodes + node_min ) -> first = arcs;

/* before below loop arc_first[i+1] is the number of arcs outgoing from i;
   after this loop arc_first[i] is the position of the first 
   outgoing from node i arcs after they would be ordered;
   this value is transformed to pointer and written to node.first[i]
   */
 
for ( i = node_min + 1; i <= node_max + 1; i ++ ) 
  {
    arc_first[i]          += arc_first[i-1];
    ( nodes + i ) -> first = arcs + arc_first[i];
  }


for ( i = node_min; i < node_max; i ++ ) /* scanning all the nodes  
                                            exept the last*/
  {

    last = ( ( nodes + i + 1 ) -> first ) - arcs;
                             /* arcs outgoing from i must be cited    
                              from position arc_first[i] to the position
                              equal to initial value of arc_first[i+1]-1  */

    for ( arc_num = arc_first[i]; arc_num < last; arc_num ++ )
      { tail = arc_tail[arc_num];

	while ( tail != i )
          /* the arc no  arc_num  is not in place because arc cited here
             must go out from i;
             we'll put it to its place and continue this process
             until an arc in this position would go out from i */

	  { arc_new_num  = arc_first[tail];
	    arc_current  = arcs + arc_num;
	    arc_new      = arcs + arc_new_num;
	    
	    /* arc_current must be cited in the position arc_new    
	       swapping these arcs:                                 */

	    head_p               = arc_new -> head;
	    arc_new -> head      = arc_current -> head;
	    arc_current -> head  = head_p;

	    cap                 = arc_new -> resCap;
	    arc_new -> resCap     = arc_current -> resCap;
	    arc_current -> resCap = cap;

	    if ( arc_new != arc_current -> rev )
	      {
	        arc_tmp                = arc_new -> rev;
	        arc_new  -> rev     = arc_current -> rev;
	        arc_current -> rev  = arc_tmp;

                ( arc_current -> rev ) -> rev = arc_current;
		( arc_new     -> rev ) -> rev = arc_new;
	      }

	    arc_tail[arc_num] = arc_tail[arc_new_num];
	    arc_tail[arc_new_num] = tail;

	    /* we increase arc_first[tail]  */
	    arc_first[tail] ++ ;

            tail = arc_tail[arc_num];
	  }
      }
    /* all arcs outgoing from  i  are in place */
  }       

/* -----------------------  arcs are ordered  ------------------------- */

/*----------- constructing lists ---------------*/


  for ( ndp = nodes + node_min; ndp <= nodes + node_max;  ndp ++ )
      ndp -> first = (arc*) NULL;

  for ( arc_current = arcs + (2*m-1); arc_current >= arcs; arc_current -- )
    {
      arc_num = arc_current - arcs;
      tail = arc_tail [arc_num];
      ndp = nodes + tail;
      /* avg
      arc_current -> next = ndp -> first;
      */
      ndp -> first = arc_current;
    }

  // for zero degree nodes
  for ( ndp = nodes + node_min; ndp <= nodes + node_max;  ndp ++ )
    if (ndp -> first == (arc*) NULL)
      ndp->first = (ndp+1)->first;

/* ----------- assigning output values ------------*/
*m_ad = m;
*n_ad = node_max - node_min + 1;
*source_ad = nodes + source;
*sink_ad   = nodes + sink;
*node_min_ad = node_min;
*nodes_ad = nodes + node_min;
*arcs_ad = arcs;
*cap_ad = acap;

for ( arc_current = arcs, arc_num = 0; 
      arc_num < 2*m;
      arc_current ++, arc_num ++
    )
     acap [ arc_num ] = arc_current -> resCap; 

if ( source < node_min || source > node_max )
  /* bad value of the source */
  { err_no = EN20; exit_on_error(err_message[err_no]); }
  
if ( (*source_ad) -> first == (arc*) NULL ||
     (*sink_ad  ) -> first == (arc*) NULL    ) 
  /* no arc goes out of the source */
  { err_no = EN20; exit_on_error(err_message[err_no]); }

/* free internal memory */
// free ( arc_first ); free ( arc_tail );

/* Thanks God! all is done */
return (0);
}
/* --------------------   end of parser  -------------------*/
