
#include <iostream>
#include <fstream>
#include <algorithm>
#include <map>

#include "tree.hpp"

using namespace std;

// naming convention: the tree has nodes and the graphs have vertices.

#include <sstream>

template <class T>
inline std::string to_string (const T& t)
{
std::stringstream ss;
ss << t;
return ss.str();
}

inline bool GHTree::in_node(int vertex, const GHTreeNode &node) 
{ 
  return node.vertices.find(vertex) != node.vertices.end(); 
}

// constructor
// creates the initial tree with respect to G with only one node
GHTree::GHTree (Edgelist& G)
{
  input_graph = &G;
  max_num_nodes = G.num_nodes();
  nodes =  new GHTreeNode[max_num_nodes];
  edges = new GHTreeEdge[2*(max_num_nodes-1)];
  nextFreeEdge = edges;
  /* the tree starts with only one node */
  num_nodes = 1;  
  GHTreeNode* v = nodes;  
  v->first_edge = 0;
  v->id = -1;

  /* the initial node contains all vertices of the tree */
  vertex_node_map.resize(max_num_nodes);
  for (int i=0; i < G.num_nodes(); i++)
    {
      v->vertices.insert(i);
      vertex_node_map[i] = v;
    }
  v->num_vertices = G.num_nodes();
  map_back = new int[max_num_nodes];
  node_map = new int[max_num_nodes];
  v->next_pivot = v->prev_pivot = 0;
  first_pivot = current_pivot = v;
  reset_task_iterators(*v);
}

GHTree::~GHTree()
{
  delete [] nodes;
  delete [] edges;
  delete [] map_back;
  delete [] node_map;
}

// return the next free pre-allocated node
GHTreeNode& GHTree::next_free_node()
{
  GHTreeNode& v = nodes[num_nodes++]; // next node
  v.first_edge = 0;
  v.id = -num_nodes;
  v.num_vertices = 0;

  return v;
}

void GHTree::reset_task_iterators(GHTreeNode& v)
{
  v.next_source = v.vertices.begin();
  v.next_sink = ++v.vertices.begin();
  v.st_step = 1;
  v.st_disp = false;
  v.st_index = 0;
  v.st_sink_reset = false;
}

// scans the subtree of the GHTree that has u as a child of the root u 
// used by build_partition(...)                    
void GHTree::scan_subtree(const GHTreeNode* u, const GHTreeNode* root, int part_number)
{
    GHTreeEdge* e;
        
    set<int>::const_iterator it;
    for (it = u->vertices.begin(); it != u->vertices.end(); ++it)
	node_map[*it] = part_number;

    e = u->first_edge;
    while (e != 0) 
      {
	/* recurse on subtrees of u */
	GHTreeNode *r = e->head;
	if (r != root) 
	  scan_subtree(r, u, part_number);
	e = e->next;
      }
}

/*  
   Build the partition with respect to the tree node v.     Vertices
   receive numbers from 0 to new_n-1, where new_n is the  number of
   vertices of the contracted graph.  Partition maps the vertices
   0..(n-1) to nodes 0..(new_n-1). The map follows the rules:
     i. vertices not in the pivot are mapped to the range 0..(part_len-1)
     ii. vertices in the pivot (noncontracted)
           are mapped to the range part_len..(part_len + pivot.vertices.size()-1)
   map_back maps the indices of the uncontracted nodes in G_contracted
   to their original indices in G.  map_back is used to change the
   tree  structure after the min-cut is computed in G_contracted. The
   contracted nodes that  correspond to the partitions are implicitly
   mapped through the formula: map_back_i = partition[i] - 1; 
   new_v is the number of nodes G_contracted will have
*/
void GHTree::build_node_map(const GHTreeNode* pivot)
{
  partition_size = 0;

  GHTreeEdge* e = pivot->first_edge;
  while (e != 0) 
    {
      /* traverse subtree */
      GHTreeNode* u = e->head;
      scan_subtree(u, pivot, partition_size++);
      e = e->next;
    };

  // treat uncontracted vertices (those in pivot)
  // numbering them from partition_size to (partition_size + 
  // + pivot->vertices.size() )
  int next_index = partition_size;
  set<int>::const_iterator it;
  for (it = pivot->vertices.begin(); it != pivot->vertices.end(); ++it)
    {
      node_map[*it] =  next_index;
      map_back[next_index] = *it;
      next_index++;
    }
}

/* print map in debugging mode */
// static void print_map(int *map, int n)
// {
//   int i;

//   printf("Map:\n");
//   for (i=0; i<n; i++)
//     printf("   %d ==> %d\n", i, map[i]);
//   printf("\n");
// }


/*
  expand_tree splits the pivot node and adds a new edge to the tree.
  edges are added to the tree exclusively inside this method.
*/
int GHTree::expand_tree(Response& msg)
{
  int i, u;
  GHTreeEdge* e;
  GHTreeEdge* e_next;
  GHTreeEdge* e_prev;

  // test if nodes s and d have been separated.
  if (vertex_node_map[msg.s] != vertex_node_map[msg.t])
    return 0;

  GHTreeNode* pivot_ptr = vertex_node_map[msg.s];
  GHTreeNode& pivot = *pivot_ptr;
  GHTreeNode& new_v = next_free_node();  // new_v is the new node

  set<int>::const_iterator it;
  for (it = pivot.vertices.begin(); it != pivot.vertices.end(); )
    {
      if (msg.cut[*it] == 1) // move *it to new_v
	{ 
	  set<int>::const_iterator it2 = it++;
     	  u = *it2;
	  vertex_node_map[u] = &new_v;
	  // add u to new_v
	  new_v.num_vertices++;
	  new_v.vertices.insert(u);
	  // remore u from pivot
	  if (*it2 == *pivot.next_source)
	    if (inc_source(&pivot))
	      {
		pivot.st_sink_reset = false;
		if (!pivot.st_disp)
		  {
		    pivot.st_disp = true;
		    for (int i = pivot.st_step; i > 0; --i)
		      inc_source(&pivot);
		  } 
	      }
	  pivot.vertices.erase(it2);
	  pivot.num_vertices--;
	} else ++it; 
    }

  /* test if the node contains only one vertex, then sets it as its id */
  if (pivot.num_vertices == 1)
    {
      pivot.id = *(pivot.vertices.begin());

      if (pivot.prev_pivot == 0) { // pivot is the first
	first_pivot = pivot.next_pivot;
      } else pivot.prev_pivot->next_pivot = pivot.next_pivot;      
      if (pivot.next_pivot != 0) {
	pivot.next_pivot->prev_pivot = pivot.prev_pivot;
	current_pivot = pivot.next_pivot;
      } else current_pivot = first_pivot;
    } else 
    {
      pivot.next_sink = pivot.next_source;
      for (int i = pivot.st_step; i > 0; --i)
	inc_sink(&pivot);
      pivot.st_index = 0;
    }

  if (new_v.num_vertices == 1)
    new_v.id = *new_v.vertices.begin();
  else {
    // insert new_v on the list of nodes with >1 vertices
    new_v.next_pivot = first_pivot;
    new_v.prev_pivot = 0;
    first_pivot = current_pivot = &new_v;
    if (new_v.next_pivot != 0)
      new_v.next_pivot->prev_pivot = &new_v;
    // reset the task iterators
    reset_task_iterators(new_v);
  }
  
  /* find neighbours to be moved */
  e = pivot.first_edge;
  e_prev = 0;
  i = 0;
  while (e != 0)
    {
      /* traverse out-edges */
      e_next = e->next; /* save next edge */
      if (msg.cut[e->head_vertex])
	{ /* move edge from pivot to new_v  */
	  e->tail = &new_v;
	  if (e_prev != 0) 
	    e_prev->next = e->next;
	  else
	    pivot.first_edge = e->next;
	  e->next = new_v.first_edge;
	  new_v.first_edge = e;
          e->reverse->head = &new_v;
	} else 
	{
	  e_prev = e;
	}
      i++;
      e = e_next;
    }; 
  
  /* create new edge between pivot and new_v */
  add_edge(new_v, pivot, msg.flow,  msg.t, msg.s); // inverted s and t

  return 1;
}

void GHTree::add_edge(GHTreeNode& u, GHTreeNode& v, double capacity, int s, int t)
{
  GHTreeEdge *e, *e_rev;

  e = nextFreeEdge++;
  e_rev = nextFreeEdge++;
  e->capacity = e_rev->capacity = capacity;
  e->tail = e_rev->head = &u;
  e->head = e_rev->tail = &v;
  e->reverse = e_rev;
  e_rev->reverse = e;

  /* add at u */
  e->next = u.first_edge;
  u.first_edge = e;
  /* add at v */
  e_rev->next = v.first_edge;
  v.first_edge = e_rev;

  e->tail_vertex = e_rev->head_vertex = s;
  e->head_vertex = e_rev->tail_vertex = t;
}

ostream& operator<<(ostream& os, GHTree& tree)
{
#ifdef DEBUG
  tree.print(true, 1);
#else
  tree.print(false, 1);
#endif
}

void GHTree::to_dot()
{
  int i;
  const GHTreeNode* u;
  const GHTreeNode* v;
  const GHTreeEdge* e;  
  map<const GHTreeNode*,string> label;

  static int tree_number = 0;

  ofstream fout;
  tree_number++;
  string fname = string("trees/t")+to_string<int>(tree_number)+".dot";
  fout.open(fname.c_str());
  
  if (!fout.good()) {
    cerr << "Error opening dot file." << endl;
    exit(1);
  }

  fout  << "graph \"grafo\" {" << endl;
  fout << "node [ color=darkseagreen ]" << endl; 

  for (i=0; i < num_nodes; i++)    
    {
      // define node labels
      u = &nodes[i];
      set<int>::const_iterator it = u->vertices.begin();
      label[u] = "\"" + to_string(*it + 1);
      for (++it; it != u->vertices.end(); ++it) 
	label[u] += "," + to_string(*it + 1);
      label[u] += "\"";
    }
  
  for (i=0; i < num_nodes; i++)    
    {
      u = &nodes[i];
      e = u->first_edge;
      while (e != 0) 
	{
	  v = e->head;
	  if (u->id < v->id)
	    fout << label[u]  << " -- " << label[v] << " [ label = \"" << e->capacity << "\" ]" << endl;
	  e = e->next;
	}
    }
  fout << "}" << endl;
  fout.close();
}

void GHTree::print (bool print_vertices, bool first_node) 
{
  int i;
  const GHTreeNode* u;
  const GHTreeNode* v;
  const GHTreeEdge* e;  

  cout << fixed;
  cout << "c ghtree" << endl;
  cout << "p max " << num_nodes << " " <<  num_nodes-1 << endl ;

  if (print_vertices)
    {
      for (i=0; i < num_nodes; i++)    
	{
	  /* print super nodes */
	  u = &nodes[i];
	  cout << "c "<< i <<  "(" << u->id + first_node << ")" << " -> ";
	  set<int>::const_iterator it;
	  for (it = u->vertices.begin(); it != u->vertices.end(); ++it) 
	    cout << *it + first_node << " ";
	  cout << endl;
	}
    }
  
  for (i=0; i < num_nodes; i++)    
    {
      u = &nodes[i];
      e = u->first_edge;
      while (e != 0) 
	{
	  v = e->head;
	  if (u->id < v->id)
	    cout << "a " << u->id + first_node << " " << v->id + first_node << " " << e->capacity << endl;
	  e = e->next;
	}
    }
}

GHTreeNode*  GHTree::pick_pivot()
{
  // return current_pivot and increment
  GHTreeNode* current = current_pivot;
  if (current_pivot->next_pivot != 0)
    current_pivot = current_pivot->next_pivot;
  else current_pivot = first_pivot;
  
  return current;
}

// generates the next task (combination of source and sink chosen from pivot)
// uses the iterators next_source and next_sink to enumerate all possible pairs within 
// the node pivot
void GHTree::pick_source_and_sink(int& source, int& sink, GHTreeNode* pivot)
{
  /* pick s and t: any two vertices in the pivot */
  source = *pivot->next_source;
  sink = *pivot->next_sink;

  // increment the iterator for the next choice of source and sink
  if (++pivot->next_sink == pivot->vertices.end())
    {
      if (++pivot->next_source == it_last(pivot->vertices))
	pivot->next_source = pivot->vertices.begin();
      pivot->next_sink = it_plus_one(pivot->next_source);
    }
}

// void GHTree::pick_source_and_sink_2(int& source, int& sink, GHTreeNode* pivot)
// {
//   /* pick s and t: any two vertices in the pivot */
//   source = *pivot->next_source;
//   sink = *pivot->next_sink;

//   // increment the iterator for the next choice of source and sink
//   ++pivot->next_source; 
//   if (pivot->next_source == pivot->vertices.end() || pivot->next_source == it_last(pivot->vertices))
//     {
//       pivot->next_source = pivot->vertices.begin();
//       pivot->next_sink = it_plus_one(pivot->next_source);
//       return ;
//     } 
//   ++pivot->next_source; 
//   if (pivot->next_source == it_last(pivot->vertices))
//     pivot->next_source = pivot->vertices.begin();
//   pivot->next_sink = it_plus_one(pivot->next_source);
// }

void GHTree::pick_source_and_sink_2(int& source, int& sink, GHTreeNode* pivot)
{
  /* pick s and t: any two vertices in the pivot */
  source = *pivot->next_source;
  sink = *pivot->next_sink;

  // increment the iterator for the next choice of source and sink
  ++pivot->next_source; 
  if (pivot->next_source == pivot->vertices.end() || pivot->next_source == it_last(pivot->vertices))
    {
      pivot->next_source = pivot->vertices.begin();
      pivot->next_sink = it_plus_one(pivot->next_source);
      return ;
    } 
  ++pivot->next_source; 
  if (pivot->next_source == it_last(pivot->vertices))
    pivot->next_source = pivot->vertices.begin();
  pivot->next_sink = it_plus_one(pivot->next_source);
}

// circularly increment the iterator next_source
// return true if the iterator was reset
bool GHTree::inc_source(GHTreeNode* pivot)
{
  ++pivot->next_source;
  if (pivot->next_source == pivot->vertices.end())
    {
      pivot->next_source = pivot->vertices.begin();
      return true;
    }
  return false;  
}

// circularly increment the iterator next_sink
bool GHTree::inc_sink(GHTreeNode* pivot)
{
  ++pivot->next_sink;
  if (pivot->next_sink == pivot->vertices.end())
    {
      pivot->next_sink = pivot->vertices.begin();
      pivot->st_sink_reset = true;
      return true;
    }
  return false;  
}

void GHTree::pick_source_and_sink_3(int& source, int& sink, GHTreeNode* pivot)
{
  int i;
  bool source_reset, sink_reset;
  /* pick s and t: any two vertices in the pivot */

  // make sure that source and sink are different vertices
  if (*pivot->next_source == *pivot->next_sink)
    inc_sink(pivot);

  source = *pivot->next_source;
  sink = *pivot->next_sink;

  source_reset = inc_source(pivot);
  inc_sink(pivot);
  ++pivot->st_index;

  if (source_reset)
    pivot->st_index = pivot->st_step;

  // test index if group ended
  if (pivot->st_index == pivot->st_step)
    {
      if (!pivot->st_sink_reset) 
	{
	  pivot->next_source = pivot->next_sink;        
	  for (  ;  pivot->st_index > 0; --pivot->st_index)
	    inc_sink(pivot);      
	} else // st_sink_reset == true
	{
	  if (!pivot->st_disp) // reset iter1 and iter2 with displacement
	    {
	      pivot->st_disp = true;
	      pivot->st_sink_reset = false;
	      pivot->next_source = pivot->vertices.begin();
	      for (i = pivot->st_step; i > 0; --i)
		inc_source(pivot);
	      pivot->next_sink = pivot->next_source;
	      for (i = pivot->st_step; i > 0; --i)
		inc_sink(pivot);
	      pivot->st_index = 0;
	    } else 
	    {
	      // increase step
	      pivot->st_step++;
	      pivot->st_disp = false;
	      pivot->st_index = 0;
	      pivot->st_sink_reset = false;
	      if (pivot->st_step <= pivot->num_vertices/2) 
		{
		  pivot->next_source = pivot->next_sink =  pivot->vertices.begin();
		  for (i = pivot->st_step; i > 0; --i)
		    inc_sink(pivot);
		} else 
		{ 
		  // start all again 
		  pivot->next_source = pivot->vertices.begin();
		  pivot->next_sink = ++pivot->vertices.begin();
		  pivot->st_step = 1;
		  pivot->st_disp = false;
		  pivot->st_index = 0;
		}
	    }
	}
    }
}

/*
  G_contracted: returned contracted graph
  map_back: a map from the vertices of the G_contracted to the nodes of the tree and of the input graph
  partition: partition numbers that each vertex belongs indexed by vertex number
  part_len: number of partitions
  n_noncontracted: number of nodes that should not be contracted
  s: chosen min-cut source
  t: chosen min-cut sink
 */

void build_contracted_graph(Edgelist& input_graph, Edgelist& G_contracted, int* node_map, int N_contracted)
{
  G_contracted.clear();
  G_contracted.set_size(N_contracted);

  /* traverse the input graph edges and add edges to the contracted graph */
  Edgelist::edge_iterator edge_it;
  int part_tail, part_head;
  for (edge_it = input_graph.begin(); edge_it != input_graph.end(); ++edge_it)
    { 
      part_tail = node_map[(*edge_it).tail];
      part_head = node_map[(*edge_it).head];
      
      // if the edge is not contained in a part of the partition
      // then add an edge to the contracted graph
      if (part_tail != part_head)  
	G_contracted.add_edge(part_tail, part_head, (*edge_it).cap);
      
      // optimize: if the edge exists, just increase its capacity
      // but it works as it is
    }
}

